#!/usr/bin/env python2
#
# Copyright (C) 2012, 2013, 2014 Ben Elliston
# Copyright (C) 2014, 2015, 2016 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Evolutionary programming applied to NEM optimisations."""

import sys
import csv
import json
import argparse
import warnings
import numpy as np

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from deap import cma
try:
    from scoop import futures
except ImportError:  # pragma: no cover
    print 'WARNING: scoop not loaded'

import nemo
from nemo import generators
from nemo import scenarios
from nemo import costs
from nemo import configfile as cf
from nemo import transmission

# Conversion factor between MWh and TWh.
twh = pow(10., 6)

# Ignore possible runtime warnings from SCOOP
warnings.simplefilter('ignore', RuntimeWarning)

parser = argparse.ArgumentParser(description='Bug reports to: nemo-devel@lists.ozlabs.org')
parser.add_argument("-c", "--carbon-price", type=int,
                    default=cf.get('costs', 'co2-price-per-t'),
                    help='carbon price ($/t) [default: %s]' % cf.get('costs', 'co2-price-per-t'))
parser.add_argument("-d", "--demand-modifier", type=str, default=[], action="append", help='demand modifier')
parser.add_argument("-g", "--generations", type=int, default=cf.get('optimiser', 'generations'),
                    help='generations [default: %s]' % cf.get('optimiser', 'generations'))
parser.add_argument("-o", "--output", type=str, default='results.json',
                    help='filename of output file (will overwrite)')
parser.add_argument("-r", "--discount-rate", type=float, default=cf.get('costs', 'discount-rate'),
                    help='discount rate [default: %s]' % cf.get('costs', 'discount-rate'))
parser.add_argument("-s", "--supply-scenario", type=str, default='re100', help='generation mix scenario [default: \'re100\']')
parser.add_argument("-t", "--transmission", action="store_true", help="include transmission [default: False]")
parser.add_argument("-v", "--verbose", action="store_true", help="be verbose")

parser.add_argument("--bioenergy-limit", type=float, default=cf.get('limits', 'bioenergy-twh-per-yr'),
                    help='Limit on annual energy from bioenergy (TWh/y) [default: %s]' %
                    cf.get('limits', 'bioenergy-twh-per-yr'))
parser.add_argument("--ccs-storage-costs", type=float, default=cf.get('costs', 'ccs-storage-costs-per-t'),
                    help='CCS storage costs ($/t) [default: %s]' % cf.get('costs', 'ccs-storage-costs-per-t'))
parser.add_argument("--coal-price", type=float, default=cf.get('costs', 'coal-price-per-gj'),
                    help='black coal price ($/GJ) [default: %s]' % cf.get('costs', 'coal-price-per-gj'))
parser.add_argument("--costs", type=str, default=cf.get('costs', 'technology-cost-class'),
                    help='cost scenario [default: %s]' % cf.get('costs', 'technology-cost-class'))
parser.add_argument("--emissions-limit", type=float, default=np.inf,
                    help='CO2 emissions limit (Mt/y) [default: infinity]')
parser.add_argument("--fossil-limit", type=float, default=1.0,
                    help='Maximum share of energy from fossil fuel [default: 1.0]')
parser.add_argument("--gas-price", type=float, default=cf.get('costs', 'gas-price-per-gj'),
                    help='gas price ($/GJ) [default: %s]' % cf.get('costs', 'gas-price-per-gj'))
parser.add_argument("--hydro-limit", type=float, default=cf.get('limits', 'hydro-twh-per-yr'),
                    help='Limit on annual energy from hydro (TWh/y) [default: %s]' %
                    cf.get('limits', 'hydro-twh-per-yr'))
parser.add_argument("--lambda", type=int, dest='lambda_', help='override CMA-ES lambda value')
parser.add_argument("--list-scenarios", action="store_true")
parser.add_argument("--min-regional-generation", type=float, default=0.0,
                    help='minimum share of energy generated intra-region [default: 0]')
parser.add_argument("--nsp-limit", type=float, default=cf.get('limits', 'nonsync-penetration'),
                    help='Non-synchronous penetration limit [default: %s]' %
                    cf.get('limits', 'nonsync-penetration'))
parser.add_argument("--reliability-std", type=float, default=0.002,
                    help='reliability standard (%% unserved)')
parser.add_argument("--reserves", type=int, default=cf.get('limits', 'minimum-reserves-mw'),
                    help='minimum operating reserves [default: %s MW]' %
                    cf.get('limits', 'minimum-reserves-mw'))
parser.add_argument("--seed", type=int, help='seed for random number generator [default: None]')
parser.add_argument("--sigma", type=float, default=cf.get('optimiser', 'sigma'),
                    help='CMA-ES sigma value [default: %s]' % cf.get('optimiser', 'sigma'))
parser.add_argument("--trace-file", type=str, help='Filename for evaluation trace (comma separated) [default: None]')
parser.add_argument('--version', action='version', version='1.0')
args = parser.parse_args()

if __name__ == '__main__' and args.list_scenarios:
    for key in sorted(scenarios.supply_scenarios):
        descr = scenarios.supply_scenarios[key].__doc__
        print '%20s' % key, '\t', descr.split('\n')[0]
    print
    sys.exit(0)

if __name__ == '__main__':
    print vars(args)

context = nemo.Context()
context.relstd = args.reliability_std

# Set the system non-synchronous penetration limit.
context.nsp_limit = args.nsp_limit
assert 0 <= context.nsp_limit <= 1

# Likewise for the minimum share of regional generation.
context.min_regional_generation = args.min_regional_generation
assert 0 <= context.min_regional_generation <= 1, \
    "Minimum regional generation must be in the interval [0,1]"

cost_class = costs.cost_switch(args.costs)
context.costs = cost_class(args.discount_rate, args.coal_price, args.gas_price, args.ccs_storage_costs)
context.costs.carbon = args.carbon_price
context.costs.transmission = transmission.Transmission(costs.txcost, args.discount_rate)
context.track_exchanges = args.transmission

# Set up the scenario.
scenarios.supply_switch(args.supply_scenario)(context)
# Apply each demand modifier argument (if any) in the given order.
for arg in args.demand_modifier:
    scenarios.demand_switch(arg)(context)

if args.verbose and __name__ == '__main__':
    docstring = scenarios.supply_switch(args.supply_scenario).__doc__
    assert docstring is not None
    # Prune off any doctest test from the docstring.
    docstring = docstring.split('\n')[0]
    print "supply scenario: %s (%s)" % (args.supply_scenario, docstring)
    print context.generators

if args.trace_file is not None:
    with open(args.trace_file, 'w') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(['# score', 'penalty', 'reasoncode', 'parameter values'])

reasons = {'unserved': 1, 'emissions': 2, 'fossil': 4, 'bioenergy': 8,
           'hydro': 16, 'reserves': 32, 'min-regional-gen': 64}


def _penalty_unserved(ctx):
    """Penalty: unserved energy"""
    minuse = ctx.total_demand() * (ctx.relstd / 100)
    use = max(0, ctx.unserved_energy() - minuse)
    reason = reasons['unserved'] if use > 0 else 0
    return pow(use, 3), reason


def _penalty_reserves(ctx):
    """Penalty: minimum reserves"""
    pen, reas = 0, 0
    for i in range(ctx.timesteps):
        reserve, spilled = 0, 0
        for g in ctx.generators:
            try:
                spilled += g.series_spilled[i]
            except KeyError:
                # non-variable generators may not have spill data
                pass

            # Calculate headroom for each generator, except pumped hydro and
            # CST -- tricky to calculate capacity
            if isinstance(g, nemo.generators.Fuelled) and not \
               isinstance(g, nemo.generators.PumpedHydro) and not \
               isinstance(g, nemo.generators.CST):
                reserve += g.capacity - g.series_power[i]

        if reserve + spilled < args.reserves:
            reas |= reasons['reserves']
            pen += pow(args.reserves - reserve + spilled, 3)
    return pen, reas


def _penalty_min_regional(ctx):
    """Penalty: minimum share of regional generation"""
    regional_generation_shortfall = 0
    for rgn in ctx.regions:
        regional_demand = 0
        for poly in rgn.polygons:
            regional_demand += ctx.demand[poly - 1].sum()
        regional_generation = 0
        for g in ctx.generators:
            if g.region() is rgn:
                regional_generation += sum(g.series_power.values())
        min_regional_generation = regional_demand * ctx.min_regional_generation
        regional_generation_shortfall += max(0, min_regional_generation - regional_generation)
    reason = reasons['min-regional-gen'] if regional_generation_shortfall > 0 else 0
    return pow(regional_generation_shortfall, 3), reason


def _penalty_emissions(ctx):
    """Penalty: total emissions"""
    emissions = 0
    for g in ctx.generators:
        if hasattr(g, 'intensity'):
            emissions += sum(g.series_power.values()) * g.intensity
    # exceedance in tonnes CO2-e
    emissions_exceedance = max(0, emissions - args.emissions_limit * pow(10, 6) * ctx.years)
    reason = reasons['emissions'] if emissions_exceedance > 0 else 0
    return pow(emissions_exceedance, 3), reason


def _penalty_fossil(ctx):
    """Penalty: limit fossil to fraction of annual demand"""
    fossil_energy = 0
    for g in ctx.generators:
        if isinstance(g, generators.Fossil):
            fossil_energy += sum(g.series_power.values())
    fossil_exceedance = max(0, fossil_energy - ctx.total_demand() * args.fossil_limit * ctx.years)
    reason = reasons['fossil'] if fossil_exceedance > 0 else 0
    return pow(fossil_exceedance, 3), reason


def _penalty_bioenergy(ctx):
    """Penalty: limit biofuel use"""
    biofuel_energy = 0
    for g in ctx.generators:
        if isinstance(g, generators.Biofuel):
            biofuel_energy += sum(g.series_power.values())
    biofuel_exceedance = max(0, biofuel_energy - args.bioenergy_limit * twh * ctx.years)
    reason = reasons['bioenergy'] if biofuel_exceedance > 0 else 0
    return pow(biofuel_exceedance, 3), reason


def _penalty_hydro(ctx):
    """Penalty: limit hydro use"""
    hydro_energy = 0
    for g in ctx.generators:
        if isinstance(g, generators.Hydro) and \
           not isinstance(g, generators.PumpedHydro):
            hydro_energy += sum(g.series_power.values())
    hydro_exceedance = max(0, hydro_energy - args.hydro_limit * twh * ctx.years)
    reason = reasons['hydro'] if hydro_exceedance > 0 else 0
    return pow(hydro_exceedance, 3), reason


# Build list of penalty functions based on command line args, etc.
penaltyfns = [_penalty_unserved, _penalty_bioenergy, _penalty_hydro]
if args.reserves > 0:
    penaltyfns.append(_penalty_reserves)
if args.emissions_limit < np.inf:
    penaltyfns.append(_penalty_emissions)
if args.fossil_limit < 1:
    penaltyfns.append(_penalty_fossil)
if context.min_regional_generation > 0:
    penaltyfns.append(_penalty_min_regional)


def cost(ctx):
    """Sum up the costs."""
    score = 0
    for g in ctx.generators:
        score += (g.capcost(ctx.costs) / ctx.costs.annuityf * ctx.years) \
            + g.opcost(ctx.costs)

    # Run through all of the penalty functions.
    penalty, reason = 0, 0
    for penaltyfn in penaltyfns:
        p, r = penaltyfn(ctx)
        penalty += p
        reason |= r

    if args.transmission:
        maxexchanges = ctx.exchanges.max(axis=0)
        np.fill_diagonal(maxexchanges, 0)
        for i in range(1, maxexchanges.shape[0]):
            # then put the max (upper, lower) into lower
            # and zero the upper entries
            for j in range(i):
                maxexchanges[i, j] = max(maxexchanges[i, j], maxexchanges[j, i])
                maxexchanges[j, i] = 0
        try:
            # existing transmission is "free"
            maxexchanges = np.maximum(0, maxexchanges - nemo.polyons.existing_net)
        except AttributeError:
            # skip if not present
            pass

        costmat = ctx.costs.transmission.cost_matrix(maxexchanges) * ctx.years
        score += costmat.sum()

    score /= ctx.total_demand()
    penalty /= ctx.total_demand()
    # Express $/yr as an average $/MWh over the period
    return score, penalty, reason


def eval_func(chromosome):
    """Average cost of energy (in $/MWh)."""
    context.set_capacities(chromosome)
    nemo.run(context)
    score, penalty, reason = cost(context)
    if args.trace_file is not None:
        # write the score and individual to the trace file
        with open(args.trace_file, 'a') as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow([score, penalty, reason] + list(chromosome))
    return score + penalty,


creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
creator.create("Individual", list, fitness=creator.FitnessMin)  # pylint: disable=no-member
toolbox = base.Toolbox()
try:
    toolbox.register("map", futures.map)
except NameError:  # pragma: no cover
    pass

# See:
# https://deap.readthedocs.org/en/master/api/algo.html#deap.cma.Strategy
# for additional parameters that can be passed to cma.Strategy.
numparams = sum(list(len(g.setters) for g in context.generators))
if args.lambda_ is None:
    # let DEAP choose
    strategy = cma.Strategy(centroid=[0] * numparams, sigma=args.sigma)
else:
    strategy = cma.Strategy(centroid=[0] * numparams, sigma=args.sigma, lambda_=args.lambda_)

toolbox.register("generate", strategy.generate, creator.Individual)  # pylint: disable=no-member
toolbox.register("update", strategy.update)
toolbox.register("evaluate", eval_func)


def run():
    """Run the evolution."""
    if args.verbose and __name__ == '__main__':
        print "objective: minimise", eval_func.__doc__

    np.random.seed(args.seed)
    hof = tools.HallOfFame(1)
    stats_fit = tools.Statistics(lambda ind: ind.fitness.values)
    stats_hof = tools.Statistics(lambda ignored: hof[0].fitness.values)
    mstats = tools.MultiStatistics(fitness=stats_fit, hallfame=stats_hof)
    mstats.register("min", np.min)

    try:
        algorithms.eaGenerateUpdate(toolbox, ngen=args.generations,
                                    stats=mstats, halloffame=hof, verbose=True)
    except KeyboardInterrupt:  # pragma: no cover
        print 'user terminated early'

    context.set_capacities(hof[0])
    nemo.run(context)
    context.verbose = True
    print
    print context
    score, penalty, reason = cost(context)
    print 'Score: %.2f $/MWh' % score
    constraints_violated = []
    if reason > 0:
        print 'Penalty: %.2f $/MWh' % penalty
        print 'Constraints violated:',
        for label, code in reasons.iteritems():
            if reason & code:
                constraints_violated += [label]
                print label,
        print
    if args.transmission:
        np.set_printoptions(precision=5)
        x = context.exchanges.max(axis=0)
        print np.array_str(x, precision=1, suppress_small=True)
        obj = {'exchanges': x.tolist(), 'generators': context}
        with open('results.json', 'w') as f:
            json.dump(obj, f, cls=nemo.Context.JSONEncoder, indent=True)

    with open(args.output, 'w') as f:
        bundle = {'options': vars(args),
                  'parameters': [max(0, cap) for cap in hof[0]],
                  'score': score, 'penalty': penalty,
                  'constraints_violated': constraints_violated}
        json.dump(bundle, f)
    print 'Done'


if __name__ == '__main__':
    run()



# Copyright (C) 2017 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""The simulation engine."""

import numpy as np
import pandas as pd
from nemo import polygons
from nemo import regions


def _sim(context, date_range):
    # reset generator internal state
    for g in context.generators:
        g.reset()

    context.exchanges.fill(0)
    generation = np.zeros((len(date_range), len(context.generators)))
    spill = np.zeros((len(date_range), len(context.generators)))

    # Extract generators in the regions of interest.
    gens = [g for g in context.generators if g.region() in context.regions]

    if context.track_exchanges:
        for g in gens:
            # every generator must be in a polygon
            assert g.polygon is not None, 'every generator must be assigned a polygon'

        selected_polygons = []
        for r in context.regions:
            if r.polygons is not None:
                selected_polygons += r.polygons

        # pull out the polygons with non-zero load in each region
        loads = [k for r in context.regions for (k, v) in r.polygons.iteritems() if v > 0]
        connections = {}
        for poly in range(1, polygons.numpolygons + 1):
            # use a list comprehension to filter the connections down to bits of interest
            connections[poly] = [path for (src, dest), path in
                                 polygons.connections.iteritems() if src is poly and
                                 dest in loads and
                                 polygons.subset(path, selected_polygons)]
            connections[poly].sort(key=polygons.pathlen)

    # Zero out polygon demands we don't care about.
    for rgn in [r for r in regions.All if r not in context.regions]:
        for poly in rgn.polygons:
            context.demand[poly - 1] = 0

    # We are free to scribble all over demand_copy.
    demand_copy = context.demand.copy()

    for hr, date in enumerate(date_range):
        hour_demand = demand_copy.loc[date].copy()
        residual_hour_demand = hour_demand.sum()
        # async_demand is the maximum amount of the demand in this
        # hour that can be met from non-synchronous
        # generation. Non-synchronous generation in excess of this
        # value must be spilled.
        async_demand = residual_hour_demand * context.nsp_limit

        if context.verbose:
            print 'STEP:', date
            print 'DEMAND:', \
                {k: round(v, 2) for k, v in hour_demand.to_dict().items()}

        # Dispatch power from each generator in merit (list) order
        for gidx, g in enumerate(gens):
            if g.non_synchronous_p and async_demand < residual_hour_demand:
                gen, spl = g.step(hr, async_demand)
            else:
                gen, spl = g.step(hr, residual_hour_demand)
            assert gen <= residual_hour_demand, \
                "generation (%.2f) > demand (%.2f) for %s" % (gen, residual_hour_demand, g)
            generation[hr, gidx] = gen

            if g.non_synchronous_p:
                async_demand -= gen
                assert async_demand > -0.1
                async_demand = max(0, async_demand)

            residual_hour_demand -= gen
            # residual can go below zero due to rounding
            assert residual_hour_demand > -0.1
            residual_hour_demand = max(0, residual_hour_demand)

            if context.verbose:
                print 'GENERATOR: %s,' % g, 'generation: %.1f' % generation[hr, gidx], \
                    'spill: %.1f' % spl, 'residual-demand: %.1f' % residual_hour_demand, \
                    'async-demand: %.1f' % async_demand

            # distribute the generation across the regions (local region first)

            if context.track_exchanges:
                paths = connections[g.polygon]
                if context.verbose:
                    print 'PATHS:', paths
                for path in paths:
                    if not gen:
                        break

                    poly = g.polygon if len(path) is 0 else path[-1][-1]
                    polyidx = poly - 1
                    transfer = gen if gen < hour_demand[polyidx] else hour_demand[polyidx]

                    if transfer > 0:
                        if context.verbose:
                            print 'DISPATCH:', int(transfer), 'to polygon', poly
                        if poly is g.polygon:
                            context.add_exchange(hr, poly, poly, transfer)
                        else:
                            # dispatch to another region
                            for src, dest in path:
                                context.add_exchange(hr, src, dest, transfer)
                                if context.verbose:
                                    print 'FLOW: polygon', src, '-> polygon', dest, '(%d)' % transfer
                                    assert polygons.direct_p(src, dest)
                        hour_demand[polyidx] -= transfer
                        gen -= transfer

            if spl > 0:
                for other in list(g for g in gens if g.storage_p):
                    stored = other.store(hr, spl)
                    spl -= stored
                    assert spl >= 0

                    # energy stored <= energy transferred, according to store's RTE
                    if context.verbose:
                        # show the energy transferred, not stored
                        print 'STORE:', g.polygon, '->', other.polygon, '(%.1f)' % stored
                    for src, dest in polygons.path(g.polygon, other.polygon):
                        context.add_exchange(hr, src, dest, stored)
            spill[hr, gidx] = spl

        if context.verbose and (hour_demand > 0).any():
            print 'RESIDUAL:', \
                {k: round(v, 2) for k, v in hour_demand.to_dict().items()}
            print 'ENDSTEP:', date

    # Change the numpy arrays to dataframes for human consumption
    context.generation = pd.DataFrame(index=date_range, data=generation)
    context.spill = pd.DataFrame(index=date_range, data=spill)

    return context


def run(context, starthour=None, endhour=None):
    """Run the simulation (without a plot).

    >>> from nemo import Context
    >>> c = Context()
    >>> c.regions = None
    >>> run(c)
    Traceback (most recent call last):
      ...
    ValueError: regions is not a list
    """
    if not isinstance(context.regions, list):
        raise ValueError('regions is not a list')

    if starthour is None:
        starthour = context.demand.index.min()
    if endhour is None:
        endhour = context.demand.index.max()
    date_range = pd.date_range(starthour, endhour, freq='H')

    _sim(context, date_range)

    # Calculate unserved energy.
    agg_demand = context.demand.sum(axis=1)
    agg_generation = context.generation.sum(axis=1)
    unserved = agg_demand - agg_generation
    # Ignore unserved events very close to 0 (rounding errors)
    context.unserved = unserved[~np.isclose(unserved, 0)]



#!/usr/bin/env python2
#
# Copyright (C) 2012, 2013, 2014 Ben Elliston
# Copyright (C) 2014, 2015 The University of New South Wales
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Replay runs from a text file of generators."""
import argparse
import json
import re
import numpy as np

import nemo
from nemo import costs
from nemo import scenarios
from nemo import utils

parser = argparse.ArgumentParser(description='Bug reports to: nemo-devel@lists.ozlabs.org')
parser.add_argument("-f", type=str, help='filename of results file (default: results.json)', metavar='FILE', required=True)
parser.add_argument("--no-legend", action="store_false", help="hide legend")
parser.add_argument("-t", "--transmission", action="store_true", help="show region exchanges [default: False]")
parser.add_argument("-v", action="count", help='verbose mode')
parser.add_argument("-x", action="store_true", help='producing a balancing plot')
parser.add_argument("--spills", action="store_true", help='plot surplus generation')
args = parser.parse_args()


def run_one(bundle):
    """Run a single simulation."""
    options = bundle['options']

    context = nemo.Context()
    context.nsp_limit = options['nsp_limit']
    assert 0 <= context.nsp_limit <= 1

    scenario = options['supply_scenario']
    scenarios.supply_switch(scenario)(context)
    print 'scenario', scenario

    # Apply each demand modifier argument (if any) in the given order.
    for arg in options['demand_modifier']:
        scenarios.demand_switch(arg)(context)

    cost_class = costs.cost_switch(options['costs'])
    context.costs = cost_class(options['discount_rate'],
                               options['coal_price'], options['gas_price'],
                               options['ccs_storage_costs'])
    context.costs.carbon = options['carbon_price']

    capacities = bundle['parameters']
    context.set_capacities(capacities)

    context.track_exchanges = args.transmission
    context.verbose = args.v > 1
    nemo.run(context)
    context.verbose = args.v > 0
    print context
    if args.transmission:
        np.set_printoptions(precision=3)
        x = context.exchanges.max(axis=0)
        print np.array_str(x, precision=1, suppress_small=True)
        with open('exchanges.json', 'w') as f:
            json.dump(x.tolist(), f)
    print

    if args.x:  # pragma: no cover
        utils.plot(context, spills=args.spills, showlegend=args.no_legend)


with open(args.f) as resultsfile:
    for line in resultsfile:
        if re.search(r'^\s*$', line):
            continue
        if re.search(r'^\s*#', line):
            print line,
            continue
        try:
            bundle = json.loads(line)
        except ValueError:
            print 'skipping malformed input:', line
            continue
        run_one(bundle)



# Copyright (C) 2017 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Utilities eg. plotting."""

import pandas as pd
import numpy as np

from matplotlib.patches import Patch
import matplotlib.pyplot as plt
from nemo import configfile
from nemo.configfile import ConfigParser
from nemo.anywh import anyWh


def _generator_list(context):
    """Return a list of the generators of interest in this run."""
    return [g for g in context.generators if g.region() in context.regions and g.capacity > 0]


def _legend(context):
    """Draw the legend."""

    # ::-1 slicing reverses the list so that the legend appears in "merit order".
    gens = _generator_list(context)[::-1]
    labels = []
    patches = []

    if len(gens) > 20:
        unique = []
        for g in gens:
            if type(g) not in unique:  # pylint: disable=unidiomatic-typecheck
                unique.append(type(g))
                # Replace the generator label with its class.
                genclass = str(type(g)).strip('<>').replace("'", "")
                labels.append(genclass.split()[1].split('.')[1])
                patches.append(g.patch)
    else:
        for g in gens:
            labels.append(g.label + ' (%s)' % anyWh(g.capacity, 'W'))
            patches.append(g.patch)

    legend = plt.figlegend([Patch('black', 'red')] + patches,
                           ['unserved'] + labels,
                           'upper right')
    plt.setp(legend.get_texts(), fontsize='small')


def plot(context, spills=False, filename=None, showlegend=True, xlim=None):
    """Produce a pretty plot of supply and demand."""
    # aggregate demand
    demand = context.demand.sum(axis=1)

    plt.clf()
    plt.ylabel('Power (MW)')
    try:
        title = configfile.get('plot', 'title')
    except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
        title = 'Supply/demand balance'
    try:
        title += '\n' + configfile.get('plot', 'subtitle')
    except (configfile.ConfigParser.NoSectionError, configfile.ConfigParser.NoOptionError):
        pass
    plt.suptitle(title)

    if showlegend:
        _legend(context)

    # Plot demand first.
    plt.plot(demand.index, demand, color='black', linewidth=3 if spills else 2)

    accum = pd.Series(data=0, index=demand.index)
    prev = accum.copy()
    for g in _generator_list(context):
        idx = context.generators.index(g)
        accum += context.generation[idx]
        # Ensure accumulated generation does not exceed demand in any timestep.
        # (Due to rounding, accum can be close to demand.)
        assert all(np.logical_or(accum < demand, np.isclose(accum, demand)))
        plt.plot(accum.index, accum, color='black', linewidth=0.5)
        plt.fill_between(accum.index, prev, accum,
                         facecolor=g.patch.get_fc(),
                         hatch=g.patch.get_hatch())
        prev = accum.copy()
    # Unmet demand is shaded red.
    plt.fill_between(accum.index, accum, demand, facecolor='red')

    if spills:
        prev = demand.copy()
        for g in list(g for g in context.generators if g.region() in context.regions):
            idx = context.generators.index(g)
            accum += context.spill[idx]
            plt.plot(accum.index, accum, color='black', linewidth=0.5)
            plt.fill_between(prev.index, prev, accum, facecolor=g.patch.get_fc(), alpha=0.3)
            prev = accum.copy()

    plt.gca().set_xlim(xlim)  # set_xlim accepts None
    plt.gca().xaxis_date()
    plt.gcf().autofmt_xdate()

    _, ymax = plt.gca().get_ylim()
    plt.plot(context.unserved.index, [ymax] * len(context.unserved),
             "yv", markersize=10, color='red', markeredgecolor='black')

    if not filename:
        plt.show()  # pragma: no cover
    else:
        plt.savefig(filename)



# Copyright (C) 2017 Ben Elliston
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Self-scaling units."""


class anyWh(object):
    """
    >>> for i in range(8): print anyWh(pow(10.23, i))
    1 MWh
    10.23 MWh
    104.65 MWh
    1.0706 GWh
    10.952 GWh
    112.04 GWh
    1.1462 TWh
    11.725 TWh
    >>> a = anyWh(500)
    >>> int(a)
    500
    >>> float(a)
    500.0
    """
    def __init__(self, n, units='Wh'):
        self._val = float(n)
        if self._val >= pow(10, 6):
            self.units = 'T' + units
        elif self._val >= pow(10, 3):
            self.units = 'G' + units
        elif self._val >= 1:
            self.units = 'M' + units
        else:
            self.units = 'k' + units

    def _scale(self):
        if self._val >= pow(10, 6):
            return self._val / pow(10, 6)
        elif self._val >= pow(10, 3):
            return self._val / pow(10, 3)
        elif self._val >= 1:
            return self._val
        return self._val * pow(10, 3)

    def __int__(self):
        return int(self._val)

    def __float__(self):
        return float(self._val)

    def __str__(self):
        return '%.5g %s' % (self._scale(), self.units)



# Copyright (C) 2010, 2011, 2014 Ben Elliston
#
# Latitude/longitude spherical geodesy formulae and scripts are
# (C) Chris Veness 2002-2011
# (www.movable-type.co.uk/scripts/latlong.html)
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

"""Latitude and longitude support for the BoM solar irradiance grids."""
import math


cellsize = 0.05
xllcorner = 112.025
yllcorner = -43.925
maxcols = 839
maxrows = 679


class LatLong(object):

    """A point of latitude and logitude."""

    def __init__(self, arg1, arg2, is_xy=False):
        """Initialise a lat/long object.

        >>> obj = LatLong(-35, 149)
        >>> obj = LatLong(1, 10, True)
        >>> obj = LatLong(1, 2, True)
        >>> obj = LatLong(679, 839, True)
        >>> obj = LatLong(839, 679, True)
        Traceback (most recent call last):
          ...
        ValueError
        >>> obj = LatLong (499, 739, True)
        >>> obj
        (-34.925, 148.975)
        """
        if is_xy:
            if arg1 > maxrows or arg2 > maxcols:
                raise ValueError
            self.lat = yllcorner + cellsize * (maxrows - arg1)
            self.lon = xllcorner + cellsize * arg2
        else:
            self.lat = arg1
            self.lon = arg2

    def xy(self):
        """
        Return the Cartesian coordinate.

        >>> obj = LatLong(-35, 149)
        >>> obj.xy()
        (499, 739)
        >>> obj = LatLong(0, 0, True)
        >>> obj
        (-9.975, 112.025)
        """
        col = int((self.lon - xllcorner) / cellsize)
        assert col < maxcols
        row = int(maxrows - ((self.lat - yllcorner) / cellsize)) - 1
        assert row >= 0
        return row, col

    def distance(self, another):
        """
        Compute the distance in kilometres between this position and another.

        >>> obj = LatLong (-35, 149)
        >>> obj2 = LatLong (-36, 150)
        >>> obj.distance (obj)
        0.0
        >>> print '%.1f' % obj.distance (obj2)
        143.4
        """
        # Code adapted from Chris Veness
        r = 6371  # km
        dlat = math.radians(another.lat - self.lat)
        dlon = math.radians(another.lon - self.lon)
        lat1 = math.radians(self.lat)
        lat2 = math.radians(another.lat)
        a = math.sin(dlat / 2) * math.sin(dlat / 2) + \
            math.sin(dlon / 2) * math.sin(dlon / 2) * \
            math.cos(lat1) * math.cos(lat2)
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        return r * c

    def __repr__(self):
        """
        Print object representation.

        >>> obj = LatLong(-35, 149)
        >>> print obj
        (-35, 149)
        """
        return self.__str__()

    def __str__(self):
        """
        Return string representation of the object.

        >>> obj = LatLong(-35, 149)
        >>> str(obj)
        '(-35, 149)'
        """
        return '(' + str(self.lat) + ', ' + str(self.lon) + ')'
